#include "mpi.h"
#include <iostream>
#include <limits>
#include <vector>
#include <cstdlib>
#include <ctime>
#include <cstring>

struct Student {
    int id;
    char name[50];
    double grade;
};

const int num_first_names = 10;
const std::string first_names[] = {
        "John", "Alice", "Bob", "Eve", "Charlie", "Grace", "Oliver", "Sophia", "David", "Lily"
};
const int num_last_names = 10;
const std::string last_names[] = {
        "Smith", "Johnson", "Brown", "Davis", "Lee", "White", "Harris", "Clark", "Young", "Lewis"
};

struct Student find_best_student(struct Student students[], int num_students) {
    struct Student best_student;
    best_student.grade = 100;
    strcpy(best_student.name, "");

    for (int i = 0; i < num_students; i++) {
        if (students[i].grade < best_student.grade ||
            (students[i].grade == best_student.grade && strcmp(students[i].name, best_student.name) < 0)) {
            best_student = students[i];
        }
    }

    return best_student;
}

MPI_Datatype create_student_mpi_type() {
    MPI_Datatype student_mpi_type;
    int block_lengths[3] = {1, 50, 1};
    MPI_Datatype types[3] = {MPI_INT, MPI_CHAR, MPI_DOUBLE};
    MPI_Aint offsets[3];

    offsets[0] = offsetof(Student, id);
    offsets[1] = offsetof(Student, name);
    offsets[2] = offsetof(Student, grade);

    MPI_Type_create_struct(3, block_lengths, offsets, types, &student_mpi_type);
    MPI_Type_commit(&student_mpi_type);

    return student_mpi_type;
}

MPI_Op create_student_min_op() {
    MPI_Op student_min;
    MPI_Op_create((MPI_User_function *) [](void *in, void *inout, int *len, MPI_Datatype *datatype) {
        Student *in_student = static_cast<Student *>(in);
        Student *inout_student = static_cast<Student *>(inout);
        if (in_student->grade > inout_student->grade) {
            *inout_student = *in_student;
        } else if (in_student->grade == inout_student->grade) {
            if (strcmp(in_student->name, inout_student->name) < 0) {
                *inout_student = *in_student;
            }
        }
    }, 1, &student_min);

    return student_min;
}

int main(int argc, char **argv) {
    MPI_Init(&argc, &argv);

    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    int num_students_per_process = 5;
    int total_students = num_students_per_process * size;

    Student *local_data = new Student[num_students_per_process];
    Student *all_students = nullptr;

    if (rank == 0) {
        all_students = new Student[total_students];
        std::cout << "Students: " << std::endl;
        srand(static_cast<unsigned>(time(0)));
        for (int i = 0; i < total_students; i++) {
            all_students[i].id = i;
            int first_name_index = rand() % num_first_names;
            int last_name_index = rand() % num_last_names;
            std::string full_name = last_names[last_name_index] + " " + first_names[first_name_index];
            strcpy(all_students[i].name, full_name.c_str());
            all_students[i].grade = rand() % 101;
            std::cout << "\tStudent - ID: "
                      << all_students[i].id << ", Name: "
                      << all_students[i].name << ", Grade: "
                      << all_students[i].grade << std::endl;
        }
    }

    MPI_Datatype student_mpi_type = create_student_mpi_type();

    MPI_Scatter(all_students, num_students_per_process, student_mpi_type, local_data, num_students_per_process,
                student_mpi_type, 0, MPI_COMM_WORLD);

    Student local_best_student;
    local_best_student.grade = std::numeric_limits<double>::max();
    local_best_student = find_best_student(local_data, num_students_per_process);

    Student global_best_student;
    global_best_student.grade = std::numeric_limits<double>::max();
    MPI_Op student_min = create_student_min_op();

    MPI_Reduce(&local_best_student, &global_best_student, 1, student_mpi_type, student_min, 0, MPI_COMM_WORLD);

    MPI_Op_free(&student_min);
    MPI_Type_free(&student_mpi_type);

    if (rank == 0) {
        std::cout << "\nBest student - ID: "
                  << global_best_student.id << ", Name: "
                  << global_best_student.name << ", Grade: "
                  << global_best_student.grade << "\n" << std::endl;
        delete[] all_students;
    }

    delete[] local_data;

    MPI_Finalize();
    return 0;
}